import base64
import binascii
import json
import math
import struct
from datetime import datetime
from typing import List, Tuple, Union

# constants used for encrypt/decrypt metadata
CONSTANTS = [1888420705, 2576816180, 2347232058, 874813317]


def raw_xxtea(v: List, n: int, k: Union[List, Tuple]) -> int:
    assert isinstance(v, list)
    assert isinstance(k, (list, tuple))
    assert isinstance(n, int)

    def mx():
        return ((z >> 5) ^ (y << 2)) + ((y >> 3) ^ (z << 4)) ^ (sum_ ^ y) + (
                    k[(p & 3) ^ e] ^ z)

    def u32(x):
        return x % 2 ** 32

    y = v[0]
    sum_ = 0
    delta = 2654435769
    if n > 1:  # Encoding
        z = v[n - 1]
        q = int(6 + (52 / n // 1))
        while q > 0:
            q -= 1
            sum_ = u32(sum_ + delta)
            e = u32(sum_ >> 2) & 3
            p = 0
            while p < n - 1:
                y = v[p + 1]
                z = v[p] = u32(v[p] + mx())
                p += 1
            y = v[0]
            z = v[n - 1] = u32(v[n - 1] + mx())
        return 0

    if n < -1:  # Decoding
        n = -n
        q = int(6 + (52 / n // 1))
        sum_ = u32(q * delta)
        while sum_ != 0:
            e = u32(sum_ >> 2) & 3
            p = n - 1
            while p > 0:
                z = v[p - 1]
                y = v[p] = u32(v[p] - mx())
                p -= 1
            z = v[n - 1]
            y = v[0] = u32(v[0] - mx())
            sum_ = u32(sum_ - delta)
        return 0
    return 1


def _bytes_to_longs(data: Union[str, bytes]) -> List[int]:
    data_bytes = data.encode() if isinstance(data, str) else data

    return [int.from_bytes(data_bytes[i:i + 4], "little")
            for i in range(0, len(data_bytes), 4)]


def _longs_to_bytes(data: List[int]) -> bytes:
    return b"".join([i.to_bytes(4, 'little') for i in data])


def _generate_hex_checksum(data: str) -> str:
    checksum = binascii.crc32(data.encode()) % 2 ** 32
    return format(checksum, "X")


class XXTEAException(Exception):
    pass


class XXTEA:
    """XXTEA wrapper class.
    
    Easy to use and compatible (by duck typing) with the Blowfish class.

    Note:    
        Partial copied from https://github.com/andersekbom/prycut and ported 
        from PY2 to PY3
    """

    def __init__(self, key: Union[str, bytes]) -> None:
        """Initializes the inner class data with the given key.

        Note:        
            The key must be 128-bit (16 characters) in length.
        """
        key = key.encode() if isinstance(key, str) else key
        if len(key) != 16:
            raise XXTEAException("Invalid key")
        self.key = struct.unpack("IIII", key)
        assert len(self.key) == 4

    def encrypt(self, data: Union[str, bytes]) -> bytes:
        """Encrypts and returns a block of data."""
        ldata = round(len(data) / 4)
        idata = _bytes_to_longs(data)
        if raw_xxtea(idata, ldata, self.key) != 0:
            raise XXTEAException("Cannot encrypt")
        return _longs_to_bytes(idata)

    def decrypt(self, data: Union[str, bytes]) -> bytes:
        """Decrypts and returns a block of data."""
        ldata = round(len(data) / 4)
        idata = _bytes_to_longs(data)
        if raw_xxtea(idata, -ldata, self.key) != 0:
            raise XXTEAException("Cannot decrypt")
        return _longs_to_bytes(idata).rstrip(b"\0")


metadata_crypter = XXTEA(_longs_to_bytes(CONSTANTS))


def encrypt_metadata(metadata: str) -> str:
    """Encrypts metadata to be used to log in to Amazon"""
    checksum = _generate_hex_checksum(metadata)
    object_str = f"{checksum}#{metadata}"
    object_enc = metadata_crypter.encrypt(object_str)
    object_base64 = base64.b64encode(object_enc).decode()

    return f"ECdITeCs:{object_base64}"


def decrypt_metadata(metadata: str) -> str:
    """Decrypts metadata for testing purposes only."""
    object_base64 = metadata.lstrip("ECdITeCs:")
    object_bytes = base64.b64decode(object_base64)
    object_dec = metadata_crypter.decrypt(object_bytes)
    object_str = object_dec.decode()
    checksum, metadata = object_str.split("#", 1)
    assert _generate_hex_checksum(metadata) == checksum

    return metadata


def now_to_unix_ms() -> int:
    return math.floor(datetime.now().timestamp() * 1000)


def meta_audible_app(user_agent: str, oauth_url: str) -> str:
    """
    Returns json-formatted metadata to simulate sign-in from iOS audible app.
    """

    meta_dict = {
        "start": now_to_unix_ms(),
        "interaction": {
            "keys": 0,
            "keyPressTimeIntervals": [],
            "copies": 0,
            "cuts": 0,
            "pastes": 0,
            "clicks": 0,
            "touches": 0,
            "mouseClickPositions": [],
            "keyCycles": [],
            "mouseCycles": [],
            "touchCycles": []
        },
        "version": "3.0.0",
        "lsUbid": "X39-6721012-8795219:1549849158",
        "timeZone": -6,
        "scripts": {
            "dynamicUrls": [
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "61HHaoAEflL._RC|11-BZEJ8lnL.js,01qkmZhGmAL.js,71qOHv6nKaL."
                 "js_.js?AUIClients/AudibleiOSMobileWhiteAuthSkin#mobile"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "21T7I7qVEeL._RC|21T1XtqIBZL.js,21WEJWRAQlL.js,31DwnWh8lFL."
                 "js,21VKEfzET-L.js,01fHQhWQYWL.js,51TfwrUQAQL.js_.js?"
                 "AUIClients/AuthenticationPortalAssets#mobile"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "0173Lf6yxEL.js?AUIClients/AuthenticationPortalInlineAssets"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "211S6hvLW6L.js?AUIClients/CVFAssets"),
                ("https://images-na.ssl-images-amazon.com/images/G/"
                 "01/x-locale/common/login/fwcim._CB454428048_.js")
            ],
            "inlineHashes": [
                -1746719145, 1334687281, -314038750, 1184642547, -137736901,
                318224283, 585973559, 1103694443, 11288800, -1611905557,
                1800521327, -1171760960, -898892073
            ],
            "elapsed": 52,
            "dynamicUrlCount": 5,
            "inlineHashesCount": 13
        },
        "plugins": "unknown||320-568-548-32-*-*-*",
        "dupedPlugins": "unknown||320-568-548-32-*-*-*",
        "screenInfo": "320-568-548-32-*-*-*",
        "capabilities": {
            "js": {
                "audio": True,
                "geolocation": True,
                "localStorage": "supported",
                "touch": True,
                "video": True,
                "webWorker": True
            },
            "css": {
                "textShadow": True,
                "textStroke": True,
                "boxShadow": True,
                "borderRadius": True,
                "borderImage": True,
                "opacity": True,
                "transform": True,
                "transition": True
            },
            "elapsed": 1
        },
        "referrer": "",
        "userAgent": user_agent,
        "location": oauth_url,
        "webDriver": None,
        "history": {
            "length": 1
        },
        "gpu": {
            "vendor": "Apple Inc.",
            "model": "Apple A9 GPU",
            "extensions": []
        },
        "math": {
            "tan": "-1.4214488238747243",
            "sin": "0.8178819121159085",
            "cos": "-0.5753861119575491"
        },
        "performance": {
            "timing": {
                "navigationStart": now_to_unix_ms(),
                "unloadEventStart": 0,
                "unloadEventEnd": 0,
                "redirectStart": 0,
                "redirectEnd": 0,
                "fetchStart": now_to_unix_ms(),
                "domainLookupStart": now_to_unix_ms(),
                "domainLookupEnd": now_to_unix_ms(),
                "connectStart": now_to_unix_ms(),
                "connectEnd": now_to_unix_ms(),
                "secureConnectionStart": now_to_unix_ms(),
                "requestStart": now_to_unix_ms(),
                "responseStart": now_to_unix_ms(),
                "responseEnd": now_to_unix_ms(),
                "domLoading": now_to_unix_ms(),
                "domInteractive": now_to_unix_ms(),
                "domContentLoadedEventStart": now_to_unix_ms(),
                "domContentLoadedEventEnd": now_to_unix_ms(),
                "domComplete": now_to_unix_ms(),
                "loadEventStart": now_to_unix_ms(),
                "loadEventEnd": now_to_unix_ms()
            }
        },
        "end": now_to_unix_ms(),
        "timeToSubmit": 108873,
        "form": {
            "email": {
                "keys": 0,
                "keyPressTimeIntervals": [],
                "copies": 0,
                "cuts": 0,
                "pastes": 0,
                "clicks": 0,
                "touches": 0,
                "mouseClickPositions": [],
                "keyCycles": [],
                "mouseCycles": [],
                "touchCycles": [],
                "width": 290,
                "height": 43,
                "checksum": "C860E86B",
                "time": 12773,
                "autocomplete": False,
                "prefilled": False
            },
            "password": {
                "keys": 0,
                "keyPressTimeIntervals": [],
                "copies": 0,
                "cuts": 0,
                "pastes": 0,
                "clicks": 0,
                "touches": 0,
                "mouseClickPositions": [],
                "keyCycles": [],
                "mouseCycles": [],
                "touchCycles": [],
                "width": 290,
                "height": 43,
                "time": 10353,
                "autocomplete": False,
                "prefilled": False
            }
        },
        "canvas": {
            "hash": -373378155,
            "emailHash": -1447130560,
            "histogramBins": []
        },
        "token": None,
        "errors": [],
        "metrics": [
            {"n": "fwcim-mercury-collector", "t": 0},
            {"n": "fwcim-instant-collector", "t": 0},
            {"n": "fwcim-element-telemetry-collector", "t": 2},
            {"n": "fwcim-script-version-collector", "t": 0},
            {"n": "fwcim-local-storage-identifier-collector", "t": 0},
            {"n": "fwcim-timezone-collector", "t": 0},
            {"n": "fwcim-script-collector", "t": 1},
            {"n": "fwcim-plugin-collector", "t": 0},
            {"n": "fwcim-capability-collector", "t": 1},
            {"n": "fwcim-browser-collector", "t": 0},
            {"n": "fwcim-history-collector", "t": 0},
            {"n": "fwcim-gpu-collector", "t": 1},
            {"n": "fwcim-battery-collector", "t": 0},
            {"n": "fwcim-dnt-collector", "t": 0},
            {"n": "fwcim-math-fingerprint-collector", "t": 0},
            {"n": "fwcim-performance-collector", "t": 0},
            {"n": "fwcim-timer-collector", "t": 0},
            {"n": "fwcim-time-to-submit-collector", "t": 0},
            {"n": "fwcim-form-input-telemetry-collector", "t": 4},
            {"n": "fwcim-canvas-collector", "t": 2},
            {"n": "fwcim-captcha-telemetry-collector", "t": 0},
            {"n": "fwcim-proof-of-work-collector", "t": 1},
            {"n": "fwcim-ubf-collector", "t": 0},
            {"n": "fwcim-timer-collector", "t": 0}
        ]
    }
    return json.dumps(meta_dict, separators=(',', ':'))



from datetime import datetime, timedelta
import io
from typing import Any, Callable, Dict, Optional
from urllib.parse import urlencode, parse_qs

from bs4 import BeautifulSoup
from PIL import Image
import httpx

from .metadata import encrypt_metadata, meta_audible_app

USER_AGENT = "Mozilla/5.0 (iPhone; CPU iPhone OS 12_3_1 like Mac OS X) " \
             "AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"


def default_captcha_callback(captcha_url: str) -> str:
    """Helper function for handling captcha."""
    # on error print captcha url instead of display captcha
    try:
        captcha = httpx.get(captcha_url).content
        f = io.BytesIO(captcha)
        img = Image.open(f)
        img.show()
    except:
        print(captcha_url)

    guess = input("Answer for CAPTCHA: ")
    return str(guess).strip().lower()


def default_otp_callback() -> str:
    """Helper function for handling 2-factor authentication."""
    guess = input("OTP Code: ")
    return str(guess).strip().lower()


def default_cvf_callback() -> str:
    """Helper function for handling cvf verifys.
    
    Amazon sends a verify code via Mail or SMS.
    """
    guess = input("CVF Code: ")
    return str(guess).strip().lower()


def default_approval_alert_callback() -> None:
    """Helper function for handling approval alerts."""
    print("Approval alert detected! Amazon sends you a mail.")
    input("Please press enter when you approve the notification.")


def get_soup(resp):
    return BeautifulSoup(resp.text, "html.parser")


def get_inputs_from_soup(soup: BeautifulSoup) -> Dict[str, str]:
    """Extracts hidden form input fields from a Amazon login page."""
    inputs = {}
    for node in soup.select("input[type=hidden]"):
        if node.attrs.get("name") and node.attrs.get("value"):
            inputs[node["name"]] = node["value"]
    return inputs


def build_oauth_url(
    country_code: str, domain: str, market_place_id: str
) -> str:
    """Builds the url to login to Amazon as an Audible device"""
    oauth_params = {
        "openid.oa2.response_type": "token",
        "openid.return_to": f"https://www.amazon.{domain}/ap/maplanding",
        "openid.assoc_handle": f"amzn_audible_ios_{country_code}",
        "openid.identity": "http://specs.openid.net/auth/2.0/"
                           "identifier_select",
        "pageId": "amzn_audible_ios",
        "accountStatusPolicy": "P1",
        "openid.claimed_id": "http://specs.openid.net/auth/2.0/"
                             "identifier_select",
        "openid.mode": "checkid_setup",
        "openid.ns.oa2": "http://www.amazon.com/ap/ext/oauth/2",
        "openid.oa2.client_id": "device:6a52316c62706d53427a57355"
                                "05a76477a45375959566674327959465"
                                "a6374424a53497069546d45234132435"
                                "a4a5a474c4b324a4a564d",
        "openid.ns.pape": "http://specs.openid.net/extensions/pape/1.0",
        "marketPlaceId": market_place_id,
        "openid.oa2.scope": "device_auth_access",
        "forceMobileLayout": "true",
        "openid.ns": "http://specs.openid.net/auth/2.0",
        "openid.pape.max_auth_age": "0"
    }

    return f"https://www.amazon.{domain}/ap/signin?{urlencode(oauth_params)}"


def check_for_captcha(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a captcha form."""
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    return True if captcha else False


def extract_captcha_url(soup: BeautifulSoup) -> Optional[str]:
    """Returns the captcha url from a Amazon login page."""
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    return captcha["src"] if captcha else None


def check_for_mfa(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a multi-factor authentication form."""
    mfa = soup.find("form", id=lambda x: x and "auth-mfa-form" in x)
    return True if mfa else False


def check_for_choice_mfa(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a MFA selection form."""
    mfa_choice = soup.find("form", id="auth-select-device-form")
    return True if mfa_choice else False


def check_for_cvf(soup: BeautifulSoup) -> bool:
    cvf = soup.find("div", id="cvf-page-content")
    return True if cvf else False


def check_for_approval_alert(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for an approval alert."""
    approval_alert = soup.find("div", id="resend-approval-alert")
    return True if approval_alert else False


def extract_cookies_from_session(session: httpx.Client) -> Dict[str, str]:
    """Extracts cookies from a httpx Client."""
    cookies = dict()
    for cookie in session.cookies.jar:
        cookies[cookie.name] = cookie.value.replace(r'"', r'')
    return cookies


def extract_token_from_url(url: httpx.URL) -> str:
    """Extracts the access token from url query after login."""
    parsed_url = parse_qs(url.query.decode())
    return parsed_url["openid.oa2.access_token"][0]


def login(
    username: str,
    password: str,
    country_code: str,
    domain: str,
    market_place_id: str,
    captcha_callback: Optional[Callable[[str], str]] = None,
    otp_callback: Optional[Callable[[], str]] = None,
    cvf_callback: Optional[Callable[[], str]] = None
) -> Dict[str, Any]:
    """Login to Audible by simulating an Audible App for iOS.
    
    Args:
        username: The Amazon email address.
        password: The Amazon password.
        country_code: The country code for the Audible marketplace to login.
        domain: domain: The top level domain for the Audible marketplace to login.
        market_place_id: The id for the Audible marketplace to login.
        captcha_callback: A custom Callable for handling captcha requests. 
            If ``None`` :func:`default_captcha_callback` is used.
        otp_callback: A custom Callable for providing one-time passwords.
            If ``None`` :func:`default_otp_callback` is used.
        cvf_callback: A custom Callback for providing the answer for a CVF code.
            If ``None`` :func:`default_cvf_callback` is used.
    
    Returns:
        An ``access_token`` with ``expires`` timestamp and the 
        ``website_cookies`` from the authorized Client.
    """

    amazon_url = f"https://www.amazon.{domain}"
    sign_in_url = amazon_url + "/ap/signin"
    cvf_url = amazon_url + "/ap/cvf/verify"

    default_headers = {"User-Agent": USER_AGENT}
    session = httpx.Client(headers=default_headers)

    while "session-token" not in session.cookies:
        session.get(amazon_url)

    oauth_url = build_oauth_url(country_code, domain, market_place_id)
    oauth_resp = session.get(oauth_url)
    oauth_soup = get_soup(oauth_resp)

    login_inputs = get_inputs_from_soup(oauth_soup)
    login_inputs["email"] = username
    login_inputs["password"] = password
    
    metadata = meta_audible_app(USER_AGENT, amazon_url)
    login_inputs["metadata1"] = encrypt_metadata(metadata)

    login_resp = session.post(sign_in_url, data=login_inputs)
    login_soup = get_soup(login_resp)

    # check for captcha
    while check_for_captcha(login_soup):
        captcha_url = extract_captcha_url(login_soup)
        if captcha_callback:
            guess = captcha_callback(captcha_url)
        else:
            guess = default_captcha_callback(captcha_url)

        inputs = get_inputs_from_soup(login_soup)
        inputs["guess"] = guess
        inputs["use_image_captcha"] = "true"
        inputs["use_audio_captcha"] = "false"
        inputs["showPasswordChecked"] = "false"
        inputs["email"] = username
        inputs["password"] = password

        login_resp = session.post(sign_in_url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for choice mfa
    # https://www.amazon.de/ap/mfa/new-otp
    while check_for_choice_mfa(login_soup):
        inputs = get_inputs_from_soup(login_soup)
        for node in login_soup.select(
            "div[data-a-input-name=otpDeviceContext]"
        ):
            # auth-TOTP, auth-SMS, auth-VOICE
            if "auth-TOTP" in node["class"]:
                inp_node = node.find("input")
                inputs[inp_node["name"]] = inp_node["value"]

        login_resp = session.post(amazon_url + "/ap/mfa", data=inputs)
        login_soup = get_soup(login_resp)

    # check for mfa (otp_code)
    while check_for_mfa(login_soup):
        if otp_callback:
            otp_code = otp_callback()
        else:
            otp_code = default_otp_callback()

        inputs = get_inputs_from_soup(login_soup)
        inputs["otpCode"] = otp_code
        inputs["mfaSubmit"] = "Submit"
        inputs["rememberDevice"] = "false"

        login_resp = session.post(sign_in_url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for cvf
    while check_for_cvf(login_soup):
        if cvf_callback:
            cvf_code = cvf_callback()
        else:
            cvf_code = default_cvf_callback()

        inputs = get_inputs_from_soup(login_soup)

        login_resp = session.post(cvf_url, data=inputs)
        login_soup = get_soup(login_resp)

        inputs = get_inputs_from_soup(login_soup)
        inputs["action"] = "code"
        inputs["code"] = cvf_code

        login_resp = session.post(cvf_url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for approval alert
    while check_for_approval_alert(login_soup):
        default_approval_alert_callback()
        url = login_soup.find_all("a", class_="a-link-normal")[1]["href"]

        login_resp = session.get(url)
        login_soup = get_soup(login_resp)

    if login_resp.status_code != 404:
        raise Exception("Unable to login")

    access_token = extract_token_from_url(login_resp.url)
    website_cookies = extract_cookies_from_session(session)
    expires = (datetime.utcnow() + timedelta(seconds=3600)).timestamp()

    session.close()

    return {
        "access_token": access_token,
        "website_cookies": website_cookies,
        "expires": expires
    }



import base64
import hashlib
import pathlib
import struct
import urllib.parse
from typing import Optional, TYPE_CHECKING, Union

import httpx

if TYPE_CHECKING:
    import audible


def get_player_id() -> str:
    """Build a software player Id."""
    player_id = base64.encodebytes(hashlib.sha1(b"").digest()).rstrip()
    return player_id.decode("ascii")


def get_player_token(auth: "audible.Authenticator") -> str:
    """Fetches a player token for further authentication.

    Args:    
        auth: The Authenticator.

    Returns:
        The player token.
    """
    params = {
        "ipRedirectOverride": True,
        "playerType": "software",
        "bp_ua": "y",
        "playerModel": "Desktop",
        "playerId": get_player_id(),
        "playerManufacturer": "Audible",
        "serial": ""
    }

    url = f"https://www.audible.{auth.locale.domain}/player-auth-token"
    with httpx.Client(cookies=auth.website_cookies) as session:
        resp = session.get(url, params=params)

    query = resp.url.query.decode()
    parsed_query = urllib.parse.parse_qs(query)
    player_token = parsed_query.get("playerToken")[0]
    return player_token


def extract_activation_bytes(data: bytes) -> str:
    """Extracts the activation bytes from activation blob.

    Args:    
        data: A activation blob returned by ``fetch_activation`` function
    
    Returns:
        The extracted activation bytes.
    
    Raises:
        ValueError: If `data` is not a valid activation blob.
    """
    if (b"BAD_LOGIN" in data or b"Whoops" in data) or \
            b"group_id" not in data:
        print(data)
        print("\nActivation failed! ;(")
        raise ValueError("data wrong")

    # https://github.com/kennedn/kindlepass/blob/master/kindlepass/kindlepass.py

    # Audible returns activation that contains metadata, extracting 0x238 bytes
    # from the end discards this metadata.
    data = data[-0x238:]

    # Strips newline characters from activation body
    fmt = "70s1x" * 8
    data = b''.join(struct.unpack(fmt, data))

    return "{:x}".format(*struct.unpack("<I", data[:4]))


def fetch_activation(player_token: str) -> bytes:
    """Fetches the activation blob from Audible server.

    Args:    
        player_token: A player token returned by ``get_player_token`` function.
    
    Returns:
        The activation blob.
    """
    url = "https://www.audible.com/license/licenseForCustomerToken"

    # register params
    rparams = {"customer_token": player_token}

    # deregister params
    dparams = {"customer_token": player_token, "action": "de-register"}

    headers = {"User-Agent": "Audible Download Manager"}
    with httpx.Client(headers=headers) as session:
        session.get(url, params=dparams)
        try:
            resp = session.get(url, params=rparams)
            return resp.content
        finally:
            session.get(url, params=dparams)


def get_activation_bytes(auth: "audible.Authenticator",
                         filename: Optional[Union[str, pathlib.Path]] = None,
                         extract: bool = True) -> Union[str, bytes]:
    """Fetches the activation blob from Audible and extracts the bytes.

    Args:    
        auth: The Authenticator.
        filename: The filename to save the activation blob (Default: None).
        extract: If True, returns the extracted activation bytes otherwise
            the whole activation blob (Default: True).
    
    Returns:
        The activation bytes or activation blob.
    """
    player_token = get_player_token(auth)
    activation = fetch_activation(player_token)
    pathlib.Path(filename).write_bytes(activation) if filename else ""

    if extract:
        activation = extract_activation_bytes(activation)

    return activation



import base64
import hmac
import json
import logging
import os
import pathlib
import re
import struct
from hashlib import sha256
from typing import Dict, Optional, Tuple, TYPE_CHECKING, Union

from pbkdf2 import PBKDF2
from pyaes import AESModeOfOperationCBC, Encrypter, Decrypter

if TYPE_CHECKING:
    import audible

logger = logging.getLogger('audible.aescipher')

BLOCK_SIZE: int = 16  # the AES block size


def aes_cbc_encrypt(key: bytes, iv: bytes, data: str,
                    padding: str = "default") -> bytes:
    """Encrypts data in cipher block chaining mode of operation.

    Args:
        key: The AES key.
        iv: The initialization vector.
        data: The data to encrypt.
        padding: Can be ``default`` or ``none`` (Default: default)

    Returns:
        The encrypted data.
    """
    encrypter = Encrypter(AESModeOfOperationCBC(key, iv), padding=padding)
    encrypted = encrypter.feed(data) + encrypter.feed()
    return encrypted


def aes_cbc_decrypt(key: bytes,
                    iv: bytes,
                    encrypted_data: bytes,
                    padding: str = "default") -> str:
    """Decrypts data encrypted in cipher block chaining mode of operation.

    Args:
        key: The AES key used at encryption.
        iv: The initialization vector used at encryption.
        encrypted_data: The encrypted data to decrypt.
        padding: Can be ``default`` or ``none`` (Default: default)
    
    Returns:
        The decrypted data.
    """
    decrypter = Decrypter(AESModeOfOperationCBC(key, iv), padding=padding)
    decrypted = decrypter.feed(encrypted_data) + decrypter.feed()
    return decrypted.decode("utf-8")


def create_salt(salt_marker: bytes,
                kdf_iterations: int) -> Tuple[bytes, bytes]:
    """Creates the header and salt for the :func:`derive_from_pbkdf2` function.

    The header consist of the number of KDF iterations encoded as a big-endian
    word bytes wrapped by ``salt_marker`` on both sides.
    The random salt has a length of 16 bytes (the AES block size) minus the
    length of the salt header.
    """
    header = (salt_marker + struct.pack('>H', kdf_iterations) + salt_marker)
    salt = os.urandom(BLOCK_SIZE - len(header))
    return header, salt


def pack_salt(header: bytes, salt: bytes) -> bytes:
    """Combines the header and salt created by :func:`create_salt` function."""
    return header + salt


def unpack_salt(packed_salt: bytes, salt_marker: bytes) -> Tuple[bytes, int]:
    """Unpack salt and kdf_iterations from previous created and packed salt."""
    mlen = len(salt_marker)
    hlen = mlen * 2 + 2

    if not (packed_salt[:mlen] == salt_marker and
            packed_salt[mlen + 2:hlen] == salt_marker):
        raise ValueError("Check salt_marker.")

    kdf_iterations = struct.unpack('>H', packed_salt[mlen:mlen + 2])[0]
    salt = packed_salt[hlen:]
    return salt, kdf_iterations


def derive_from_pbkdf2(password: str,
                       *,
                       key_size: int,
                       salt: bytes,
                       kdf_iterations: int,
                       hashmod,
                       mac) -> bytes:
    """Creates an AES key with the :class:`PBKDF2` key derivation class."""
    kdf = PBKDF2(password, salt, min(kdf_iterations, 65535), hashmod, mac)
    return kdf.read(key_size)


class AESCipher:
    """Encrypt/Decrypt data using password to generate key.

    The encryption algorithm used is symmetric AES in cipher-block chaining
    (CBC) mode.

    The key is derived via the PBKDF2 key derivation function (KDF) from the
    password and a random salt of 16 bytes (the AES block size) minus the
    length of the salt header (see below).
    The hash function used by PBKDF2 is SHA256 per default. You can pass a
    different hash function module via the ``hashmod`` argument. The module
    must adhere to the Python API for Cryptographic Hash Functions (PEP 247).
    PBKDF2 uses a number of iterations of the hash function to derive the key,
    which can be set via the ``kdf_iterations`` keyword argument. The default
    number is 1000 and the maximum 65535.
    The header and the salt are written to the first block of the encrypted
    output (bytes mode) or written as key/value pairs (dict mode). The header
    consist of the number of KDF iterations encoded as a big-endian word bytes
    wrapped by ``salt_marker`` on both sides. With the default value of
    ``salt_marker = b'$'``, the header size is thus 4 and the salt 12 bytes.
    The salt marker must be a byte string of 1-6 bytes length.
    The last block of the encrypted output is padded with up to 16 bytes, all
    having the value of the length of the padding.
    All values in dict mode are written as base64 encoded string.

    Attributes:
        password: The password for encryption/decryption.
        key_size: The size of the key. Can be ``16``, ``24`` or ``32``
            (Default: 32).
        salt_marker: The salt marker with max. length of 6 bytes (Default: $).
        kdf_iterations: The number of iterations of the hash function to
            derive the key (Default: 1000).
        hashmod: The hash method to use (Default: sha256).
        mac: The mac module to use (Default: hmac).

    Args:
        password: The password for encryption/decryption.
        key_size: The size of the key. Can be ``16``, ``24`` or ``32``
            (Default: 32).
        salt_marker: The salt marker with max. length of 6 bytes (Default: $).
        kdf_iterations: The number of iterations of the hash function to
            derive the key (Default: 1000).
        hashmod: The hash method to use (Default: sha256).
        mac: The mac module to use (Default: hmac).
    
    Raises:
        ValueError: If `salt_marker` is not one to six bytes long.
        ValueError: If `kdf_iterations` is greater than 65535.
        TypeError: If type of `salt_marker` is not bytes.
    """

    def __init__(self,
                 password: str,
                 *,
                 key_size: int = 32,
                 salt_marker: bytes = b"$",
                 kdf_iterations: int = 1000,
                 hashmod=sha256,
                 mac=hmac) -> None:

        if not 1 <= len(salt_marker) <= 6:
            raise ValueError('The salt_marker must be one to six bytes long.')

        if not isinstance(salt_marker, bytes):
            raise TypeError('salt_marker must be a bytes instance.')

        if kdf_iterations >= 65536:
            raise ValueError('kdf_iterations must be <= 65535.')

        self.password = password
        self.key_size = key_size
        self.hashmod = hashmod
        self.mac = mac
        self.salt_marker = salt_marker
        self.kdf_iterations = kdf_iterations

    def _encrypt(self, data: str) -> Tuple[bytes, bytes, bytes]:
        header, salt = create_salt(self.salt_marker, self.kdf_iterations)
        key = derive_from_pbkdf2(
            password=self.password,
            key_size=self.key_size,
            salt=salt,
            kdf_iterations=self.kdf_iterations,
            hashmod=self.hashmod,
            mac=self.mac)
        iv = os.urandom(BLOCK_SIZE)
        encrypted_data = aes_cbc_encrypt(key, iv, data)
        return pack_salt(header, salt), iv, encrypted_data

    def _decrypt(self, salt: bytes, iv: bytes, encrypted_data: bytes) -> str:
        try:
            salt, kdf_iterations = unpack_salt(salt, self.salt_marker)
        except ValueError:
            kdf_iterations = self.kdf_iterations

        key = derive_from_pbkdf2(
            password=self.password,
            key_size=self.key_size,
            salt=salt,
            kdf_iterations=kdf_iterations,
            hashmod=self.hashmod,
            mac=self.mac)
        return aes_cbc_decrypt(key, iv, encrypted_data)

    def to_dict(self, data: str) -> Dict[str, str]:
        """Encrypts data in dict style.

        The output dict contains the base64 encoded (packed) salt, iv and
        ciphertext key/value pairs and an info key/value pair with additional
        encryption information.

        Args:
            data: The data to encrypt.
        
        Returns:
            The encrypted data in dict style.
        """
        salt, iv, encrypted_data = self._encrypt(data)

        return {
            "salt": base64.b64encode(salt).decode("utf-8"),
            "iv": base64.b64encode(iv).decode("utf-8"),
            "ciphertext": base64.b64encode(encrypted_data).decode("utf-8"),
            "info": 'base64-encoded AES-CBC-256 of JSON object'
        }

    def from_dict(self, data: dict) -> str:
        """Decrypts data previously encrypted with :meth:`AESCipher.to_dict`.

        Args:
            data: The encrypted data in json style.
        
        Returns:
            The decrypted data.
        """
        salt = base64.b64decode(data["salt"])
        iv = base64.b64decode(data["iv"])
        encrypted_data = base64.b64decode(data["ciphertext"])
        return self._decrypt(salt, iv, encrypted_data)

    def to_bytes(self, data: str) -> bytes:
        """Encrypts data in bytes style.

        The output bytes contains the (packed) salt, iv and ciphertext.

        Args:
            data: The data to encrypt.
        
        Returns:
            The encrypted data in dict style.
        """
        salt, iv, encrypted_data = self._encrypt(data)
        return salt + iv + encrypted_data

    def from_bytes(self, data: bytes) -> str:
        """Decrypts data previously encrypted with :meth:`AESCipher.to_bytes`.

        Args:
            data: The encrypted data in bytes style.
        
        Returns:
            The decrypted data.
        """
        bs = BLOCK_SIZE
        salt = data[:bs]
        iv = data[bs:2 * bs]
        encrypted_data = data[2 * bs:]
        return self._decrypt(salt, iv, encrypted_data)

    def to_file(self,
                data: str,
                filename: pathlib.Path,
                encryption: str = "json",
                indent: int = 4) -> None:
        """Encrypts and saves data to given file.

        Args:
            data: The data to encrypt.
            filename: The name of the file to save the data to.
            encryption: The encryption style to use. Can be ``json`` or 
                ``bytes`` (Default: json).
            indent: The indention level when saving in json style
                (Default: 4).
        
        Raises:
            ValueError: If `encryption` is not ``json`` or ``bytes``.
        """
        if encryption == "json":
            encrypted_dict = self.to_dict(data)
            data_json = json.dumps(encrypted_dict, indent=indent)
            filename.write_text(data_json)

        elif encryption == "bytes":
            encrypted_data = self.to_bytes(data)
            filename.write_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\"..")

    def from_file(self, filename: pathlib.Path,
                  encryption: str = "json") -> str:
        """Loads and decrypts data from given file.

        Args:
            filename: The name of the file to load the data from.
            encryption: The encryption style which where used. Can be ``json`` 
                or ``bytes`` (Default: json).

        Returns:
            The decrypted data.
        
        Raises:
            ValueError: If `encryption` is not ``json`` or ``bytes``.        
        """
        if encryption == "json":
            encrypted_json = filename.read_text()
            encrypted_dict = json.loads(encrypted_json)
            return self.from_dict(encrypted_dict)

        elif encryption == "bytes":
            encrypted_data = filename.read_bytes()
            return self.from_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\".")


def detect_file_encryption(filename: pathlib.Path) -> Optional[str]:
    """Detect the encryption format from an authentication file.

    Args:
        filename: The name for the authentication file.

    Returns:
        ``False`` if file is not encrypted otherwise the encryption format.
    """
    file = filename.read_bytes()
    encryption = None

    try:
        file = json.loads(file)
        if "adp_token" in file:
            encryption = False
        elif "ciphertext" in file:
            encryption = "json"
    except UnicodeDecodeError:
        encryption = "bytes"

    return encryption


def remove_file_encryption(source: Union[str, pathlib.Path],
                           target: Union[str, pathlib.Path],
                           password: str,
                           **kwargs) -> None:
    """Removes the encryption from an authentication file.

    Please try to load the authentication file with
    :meth:`audible.Authenticator.from_file` and save the authentication data
    as a unencrypted file first. Use this function as fallback if you ran into
    any error.

    Args:
        source: The encrypted authentication file.
        target: The filename for the decrypted file.
        password: The password for the encrypted authentication file.

    Raises:
        ValueError: If ``source`` is not encrypted.
    """
    source_file = pathlib.Path(source)
    encryption = detect_file_encryption(source_file)

    if not encryption:
        raise ValueError("file is not encrypted")

    crypter = AESCipher(password, **kwargs)
    decrypted = crypter.from_file(source_file, encryption=encryption)
    pathlib.Path(target).write_text(decrypted)


def _decrypt_voucher(device_serial_number: str,
                     customer_id: str,
                     device_type: str,
                     asin: str,
                     voucher: Dict) -> Dict:
    # https://github.com/mkb79/Audible/issues/3#issuecomment-705262614
    buf = (device_type + device_serial_number + customer_id +
           asin).encode("ascii")
    digest = sha256(buf).digest()
    key = digest[0:16]
    iv = digest[16:]

    # decrypt "voucher" using AES in CBC mode with no padding
    b64d_voucher = base64.b64decode(voucher)
    plaintext = aes_cbc_decrypt(
        key, iv, b64d_voucher, padding="none").rstrip("\x00")

    try:
        return json.loads(plaintext)
    except json.JSONDecodeError:
        fmt = r"^{\"key\":\"(?P<key>.*?)\",\"iv\":\"(?P<iv>.*?)\","
        match = re.match(fmt, plaintext)
        return match.groupdict()


def decrypt_voucher_from_licenserequest(auth: "audible.Authenticator",
                                        license_response: Dict) -> Dict:
    """Decrypt the voucher from license request response.

    Args:
        auth: The Authenticator.
        license_response: The response content from a
            :http:post:`/1.0/content/(string:asin)/licenserequest` request.
    
    Returns:
        The decrypted license voucher with needed key and iv.

    Note:
        A device registration is needed to use the auth instance for a
        license request and to obtain the needed device data
    """
    # device data
    device_info = auth.device_info
    device_serial_number = device_info["device_serial_number"]
    device_type = device_info["device_type"]

    # user data
    customer_id = auth.customer_info["user_id"]

    # book specific data
    asin = license_response["content_license"]["asin"]
    encrypted_voucher = license_response["content_license"]["license_response"]

    return _decrypt_voucher(
        device_serial_number=device_serial_number,
        customer_id=customer_id,
        device_type=device_type,
        asin=asin,
        voucher=encrypted_voucher)



import uuid
from datetime import datetime, timedelta
from typing import Any, Dict

import httpx


def get_random_device_serial() -> str:
    """Generates and prepares a random uuid version 4.
    
    Using a random serial prevents from unregister devices by other users
    with same `device_serial`.
    """
    return str(uuid.uuid4()).replace("-", "")


def register(access_token: str, domain: str) -> Dict[str, Any]:
    """Registers a dummy Audible device. 

    Args:
        access_token: An access token fetches from :func:`audible.auth.login`.
        domain: The top level domain of the requested Amazon server (e.g. com).
    
    Returns:
        Additional authentication data needed for access Audible API.
    """
    body = {
        "requested_token_type":
            ["bearer", "mac_dms", "website_cookies",
             "store_authentication_cookie"],
        "cookies": {
            "website_cookies": [],
            "domain": f".amazon.{domain}"},
        "registration_data": {
            "domain":
                "Device",
            "app_version":
                "3.26.1",
            "device_serial":
                get_random_device_serial(),
            "device_type":
                "A2CZJZGLK2JJVM",
            "device_name": (
                "%FIRST_NAME%%FIRST_NAME_POSSESSIVE_STRING%%DUPE_"
                "STRATEGY_1ST%Audible for iPhone"),
            "os_version":
                "13.5.1",
            "device_model":
                "iPhone",
            "app_name":
                "Audible"},
        "auth_data": {
            "access_token": access_token},
        "requested_extensions": ["device_info", "customer_info"]
    }

    resp = httpx.post(f"https://api.amazon.{domain}/auth/register", json=body)

    resp_json = resp.json()
    if resp.status_code != 200:
        raise Exception(resp_json)

    success_response = resp_json["response"]["success"]

    tokens = success_response["tokens"]
    adp_token = tokens["mac_dms"]["adp_token"]
    device_private_key = tokens["mac_dms"]["device_private_key"]
    store_authentication_cookie = tokens["store_authentication_cookie"]
    access_token = tokens["bearer"]["access_token"]
    refresh_token = tokens["bearer"]["refresh_token"]
    expires_s = int(tokens["bearer"]["expires_in"])
    expires = (datetime.utcnow() + timedelta(seconds=expires_s)).timestamp()

    extensions = success_response["extensions"]
    device_info = extensions["device_info"]
    customer_info = extensions["customer_info"]

    website_cookies = dict()
    for cookie in tokens["website_cookies"]:
        website_cookies[cookie["Name"]] = cookie["Value"].replace(r'"', r'')

    return {
        "adp_token": adp_token,
        "device_private_key": device_private_key,
        "access_token": access_token,
        "refresh_token": refresh_token,
        "expires": expires,
        "website_cookies": website_cookies,
        "store_authentication_cookie": store_authentication_cookie,
        "device_info": device_info,
        "customer_info": customer_info
    }


def deregister(access_token: str, domain: str,
               deregister_all: bool = False) -> Dict[str, Any]:
    """Deregisters a dummy Audible device.
    
    Note:
        Except of the ``access_token``, all authentication data will loose 
        validation immediately.

    Args:
        access_token: The access token from the previous registered device 
            which you want to deregister.
        domain: The top level domain of the requested Amazon server (e.g. com).
        deregister_all: If ``True``, deregister all Audible devices on Amazon.

    Returns:
        The response for the deregister request. Contains errors, if some occurs.
    """
    body = {"deregister_all_existing_accounts": deregister_all}
    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.post(
        f"https://api.amazon.{domain}/auth/deregister",
        json=body,
        headers=headers
    )

    resp_json = resp.json()
    if not resp.status_code == 200:
        raise Exception(resp_json)

    return resp_json



import logging
from typing import Dict, Optional
from urllib.parse import parse_qs, urlparse

import httpx
from bs4 import BeautifulSoup
from httpcore import ConnectError

logger = logging.getLogger("audible.localization")

LOCALE_TEMPLATES = {
    "germany": {
        "country_code": "de",
        "domain": "de",
        "market_place_id": "AN7V1F1VY261K"
    },
    "united_states": {
        "country_code": "us",
        "domain": "com",
        "market_place_id": "AF2M0KC94RCEA"
    },
    "united_kingdom": {
        "country_code": "uk",
        "domain": "co.uk",
        "market_place_id": "A2I9A3Q2GNFNGQ"
    },
    "france": {
        "country_code": "fr",
        "domain": "fr",
        "market_place_id": "A2728XDNODOQ8T"
    },
    "canada": {
        "country_code": "ca",
        "domain": "ca",
        "market_place_id": "A2CQZ5RBY40XE"
    },
    "italy": {
        "country_code": "it",
        "domain": "it",
        "market_place_id": "A2N7FU2W2BU2ZC"
    },
    "australia": {
        "country_code": "au",
        "domain": "com.au",
        "market_place_id": "AN7EY7DTAW63G"
    },
    "india": {
        "country_code": "in",
        "domain": "in",
        "market_place_id": "AJO3FBRUE6J4S"
    },
    "japan": {
        "country_code": "jp",
        "domain": "co.jp",
        "market_place_id": "A1QAP3MOU4173J"
    }
}


def search_template(key: str, value: str) -> Optional[Dict[str, str]]:
    for country in LOCALE_TEMPLATES:
        locale = LOCALE_TEMPLATES[country]
        if locale[key] == value:
            logger.debug(f"found locale for {country}")
            return locale

    logger.info(f"don\'t found {value} in {key}")
    return None


def autodetect_locale(domain: str) -> Dict[str, str]:
    """Try to automatically detect correct settings for marketplace.

    Needs the top level domain of the audible page to continue with
    (e.g. co.uk, co.jp) and returns results found.
    
    Args:
        domain: The top level domain for the Audible marketplace to
            detect settings for (e.g. com).
    
    Returns:
        The settings for the found Audible marketplace.
    """
    domain = domain.lstrip(".")
    site = f"https://www.audible.{domain}"
    params = {"ipRedirectOverride": True, "overrideBaseCountry": True}

    try:
        resp = httpx.get(site, params=params)
    except ConnectError as e:
        logger.warning(f"site {site} doesn\'t exists or Network Error occours")
        raise e

    soup = BeautifulSoup(resp.text, "html.parser")

    login_link = soup.find("a", class_="ui-it-sign-in-link")["href"]
    parsed_link = urlparse(login_link)
    query_string = parse_qs(parsed_link.query)
    market_place_id = query_string['marketPlaceId'][0]
    country_code = query_string['pageId'][0].split("_")[-1]

    return {
        "country_code": country_code,
        "domain": domain,
        "market_place_id": market_place_id
    }


class Locale:
    """
    Adjustments for the different marketplaces who are provided by Audible.

    Look at `locales.json` for examples. You can try to
    ``autodetect_locale`` if your marketplace is not in `locales.json`.
    
    """

    def __init__(
            self,
            country_code: Optional[str] = None,
            domain: Optional[str] = None,
            market_place_id: Optional[str] = None
    ) -> None:

        if not all([country_code, domain, market_place_id]):
            locale = None
            if country_code:
                locale = search_template("country_code", country_code)
            elif domain:
                locale = search_template("domain", domain)

            if locale is None:
                raise Exception("can\'t find locale")

        self._country_code = country_code or locale["country_code"]
        self._domain = domain or locale["domain"]
        self._market_place_id = market_place_id or locale["market_place_id"]

    def __repr__(self):
        return (
            f"Locale class for domain: {self.domain}, "
            f"marketplace: {self.market_place_id}"
        )

    def to_dict(self) -> Dict[str, str]:
        return {
            "country_code": self.country_code,
            "domain": self.domain,
            "market_place_id": self.market_place_id
        }

    @property
    def country_code(self) -> str:
        return self._country_code

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def market_place_id(self) -> str:
        return self._market_place_id



__title__ = "audible"
__description__ = "A(Sync) Interface for internal Audible API written in " \
                  "pure Python."
__url__ = "https://github.com/mkb79/audible"
__version__ = "0.5.0"
__author__ = "mkb79"
__author_email__ = "mkb79@hackitall.de"
__license__ = "AGPL"
__status__ = "Development"



import logging
import pathlib
import re
import time
from typing import Any, Dict, Optional, Union

from .aescipher import AESCipher
from .localization import Locale


logger = logging.getLogger("audible.utils")


def _check_website_cookies(value: Dict[str, str]) -> None:
    if not isinstance(value, dict):
        raise TypeError(f"website_cookies: Expected dict, "
                        f"got {type(value).__name__}.")
    if not set(map(type, value.values())) == {str}:
        raise TypeError("website_cookies: Value(s) of website_cookies have "
                        "wrong type.")


def _check_adp_token(value: str) -> None:
    if not isinstance(value, str):
        raise TypeError(f"adp_token: Expected str, "
                        f"got {type(value).__name__}.")

    fmt = (r"^{enc:(?P<enc>.*?)}{key:(?P<key>.*?)}{iv:(?P<iv>.*?)}"
           r"{name:(?P<name>.*?)}{serial:(?P<serial>Mg==)}$")
    if not re.match(fmt, value):
        raise ValueError("adp_token: Invalid token.")


def _check_access_token(value: str) -> None:
    if not isinstance(value, str):
        raise TypeError(f"access_token: Expected str, "
                        f"got {type(value).__name__}.")

    fmt = r"^(?P<access_token>Atna\|.*)$"
    if not re.match(fmt, value):
        raise ValueError("access_token: Invalid token.")


def _check_refresh_token(value: str) -> None:
    if not isinstance(value, str):
        raise TypeError(f"refresh_token: Expected str, "
                        f"got {type(value).__name__}.")

    fmt = r"^(?P<refresh_token>Atnr\|.*)$"
    if not re.match(fmt, value):
        raise ValueError("refresh_token: Invalid token.")


def _check_device_private_key(value: str) -> None:
    if not isinstance(value, str):
        raise TypeError(f"device_private_key: Expected str, "
                        f"got {type(value).__name__}.")

    fmt = (r"^(?P<device_private_key>-----BEGIN RSA PRIVATE KEY-----.*"
           r"-----END RSA PRIVATE KEY-----\n)$")
    if not re.match(fmt, value, re.S):
        raise ValueError("device_private_key: Invalid token.")


def _check_expires(value: Union[int, float, str]) -> Optional[float]:
    if not isinstance(value, (int, float, str)):
        raise TypeError(f"expires: Expected int/float/str, "
                        f"got {type(value).__name__}.")

    if isinstance(value, str):
        try:
            return float(value)
        except ValueError as exc:
            raise ValueError("expires: Got str. Converting to float "
                             "raises an error.") from exc


def _check_locale(value: Union[str, Locale]) -> Optional[Locale]:
    if not isinstance(value, (Locale, str)):
        raise TypeError(f"locales: Expected Locale/str, "
                        f"got {type(value).__name__}.")

    if isinstance(value, str):
        return Locale(value.lower())


def _check_filename(value) -> pathlib.Path:
    if not isinstance(value, pathlib.Path):
        try:
            return pathlib.Path(value)
        except Exception as exc:
            raise Exception(f"filename: Got {type(value).__name__}. Converting "
                            f"to Path raises an error.") from exc


def _check_crypter(value: AESCipher) -> None:
    if not isinstance(value, AESCipher):
        raise TypeError(f"crypter: Expected AESCipher, "
                        f"got {type(value).__name__}.")


def _check_encryption(value: Union[bool, str]) -> None:
    if not isinstance(value, (bool, str)):
        raise TypeError(f"encryption: Expected bool/str, "
                        f"got {type(value).__name__}.")

    if value not in [False, "json", "bytes"]:
        raise ValueError("encryption: Value are not allowed.")


string_function_map = {
    "website_cookies": _check_website_cookies,
    "adp_token": _check_adp_token,
    "access_token": _check_access_token,
    "refresh_token": _check_refresh_token,
    "device_private_key": _check_device_private_key,
    "expires": _check_expires,
    "locale": _check_locale,
    "filename": _check_filename,
    "crypter": _check_crypter,
    "encryption": _check_encryption
}


def test_convert(key: str, value: Any) -> Any:
    """Helper function to check and convert values for specific keys."""
    if key in string_function_map and value is not None:
        return string_function_map[key](value) or value

    return value


class ElapsedTime:
    def __init__(self):
        self.start_time = time.time()

    def __call__(self):
        return time.time() - self.start_time



import json
import logging
from typing import Dict, Optional, Tuple, Union

import httpx
from httpx import URL

from .auth import Authenticator
from .exceptions import (
    BadRequest, NotFoundError, NotResponding, NetworkError, ServerError,
    Unauthorized, UnexpectedError, RatelimitError, RequestError
)
from .localization import LOCALE_TEMPLATES, Locale

logger = logging.getLogger("audible.client")


def convert_response_content(resp: httpx.Response) -> Union[Dict, str]:
    try:
        return resp.json()
    except json.JSONDecodeError:
        return resp.text


class Client:
    _API_URL_TEMP = "https://api.audible."
    _API_VERSION = "1.0"
    _SESSION = httpx.Client
    _REQUEST_LOG = "{method} {url} has received {text}, has returned {status}"

    def __init__(
            self,
            auth: Authenticator,
            country_code: Optional[str] = None,
            timeout: int = 10
    ):
        locale = Locale(country_code.lower()) if country_code else auth.locale
        api_url = self._API_URL_TEMP + locale.domain
        headers = {
            "Accept": "application/json",
            "Accept-Charset": "utf-8",
            "Content-Type": "application/json"
        }
        self.session = self._SESSION(
            headers=headers, timeout=timeout, base_url=api_url, auth=auth
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        return f"<Sync Client for *{self.marketplace}* marketplace>"

    def close(self) -> None:
        self.session.close()

    def switch_marketplace(self, country_code: str) -> None:
        locale = Locale(country_code.lower())
        self.session.base_url = URL(self._API_URL_TEMP + locale.domain)

    @property
    def marketplace(self) -> str:
        base_url = str(self.session.base_url)
        slice_len = len(self._API_URL_TEMP)
        domain = base_url[slice_len:]

        for country in LOCALE_TEMPLATES.values():
            if domain == country["domain"]:
                return country["country_code"]

    @property
    def auth(self) -> Authenticator:
        return self.session.auth

    def switch_user(
            self,
            auth: Authenticator,
            switch_to_default_marketplace: bool = False
    ) -> None:
        if switch_to_default_marketplace:
            self.switch_marketplace(auth.locale.country_code)
        self.session.auth = auth

    def get_user_profile(self) -> Dict:
        self.auth.refresh_access_token()
        return self.auth.user_profile()

    @property
    def user_name(self) -> str:
        user_profile = self.get_user_profile()
        return user_profile["name"]

    def _raise_for_status_error(self, resp: httpx.Response):
        code = resp.status_code

        data = convert_response_content(resp)

        if 300 > code >= 200:  # Request was successful
            return
        elif code == 400:
            raise BadRequest(resp, data)
        elif code in (401, 403):  # Unauthorized request - Invalid credentials
            raise Unauthorized(resp, data)
        elif code == 404:  # not found
            raise NotFoundError(resp, data)
        elif code == 429:
            raise RatelimitError(resp, data)
        elif code == 503:  # Maintainence
            raise ServerError(resp, data)
        else:
            raise UnexpectedError(resp, data)

    def _prepare_path(self, path: str) -> str:
        if path.startswith("/"):
            path = path[1:]

        if path.startswith(self._API_VERSION):
            return path

        return "/".join((self._API_VERSION, path))

    def _request(self, method: str, path: str, **kwargs) -> Union[Dict, str]:
        url = self._prepare_path(path)

        try:
            resp = self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method,
                    url=resp.url,
                    text=resp.text,
                    status=resp.status_code
                )
            )

            resp.raise_for_status()

            return convert_response_content(resp)

        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        except httpx.RequestError as exc:
            raise RequestError(exc)
        except httpx.HTTPStatusError as exc:
            self._raise_for_status_error(exc.response)

        finally:
            try:
                resp.close()
            except UnboundLocalError:
                pass

    def _split_kwargs(self, **kwargs) -> Tuple:
        protected_kwargs = [
            "method", "url", "params", "data", "json", "headers", "cookies",
            "files", "auth", "timeout", "allow_redirects", "proxies", "verify",
            "stream", "cert"
        ]
        params = kwargs.pop("params", {})
        for key in list(kwargs.keys()):
            if key not in protected_kwargs:
                params[key] = kwargs.pop(key)

        return params, kwargs

    def get(self, path: str, **kwargs) -> Union[Dict, str]:
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("GET", path, params=params, **kwargs)

    def post(self, path: str, body: Dict, **kwargs) -> Union[Dict, str]:
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("POST", path, params=params, json=body, **kwargs)

    def delete(self, path: str, **kwargs) -> Union[Dict, str]:
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("DELETE", path, params=params, **kwargs)

    def put(self, path: str, body: Dict, **kwargs) -> Union[Dict, str]:
        params, kwargs = self._split_kwargs(**kwargs)
        return self._request("PUT", path, params=params, json=body, **kwargs)


class AsyncClient(Client):
    _SESSION = httpx.AsyncClient

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    def __repr__(self):
        return f"<AyncClient for *{self.marketplace}* marketplace>"

    async def close(self) -> None:
        await self.session.aclose()

    async def _request(self, method: str, path: str, **kwargs) -> Union[Dict, str]:
        url = self._prepare_path(path)

        try:
            resp = await self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method,
                    url=resp.url,
                    text=resp.text,
                    status=resp.status_code
                )
            )

            resp.raise_for_status()

            return convert_response_content(resp)

        except (
                httpx.ConnectTimeout, httpx.ReadTimeout, httpx.WriteTimeout,
                httpx.PoolTimeout
        ):
            raise NotResponding
        except httpx.NetworkError:
            raise NetworkError
        except httpx.RequestError as exc:
            raise RequestError(exc)
        except httpx.HTTPStatusError as exc:
            self._raise_for_status_error(exc.response)

        finally:
            try:
                await resp.aclose()
            except UnboundLocalError:
                pass



class RequestError(Exception):
    """Base class for all errors"""


class StatusError(RequestError):
    """
    Base class for all errors except NotResponding and
    RatelimitDetectedError
    """

    def __init__(self, resp, data):
        self.response = resp
        self.code = getattr(resp, 'status_code')
        self.method = getattr(resp, 'method', None)
        self.reason = resp.reason_phrase
        if isinstance(data, dict):
            self.error = data.get('error')
            if 'message' in data:
                self.error = data.get('message')
        else:
            self.error = data
        self.fmt = '{0.reason} ({0.code}): {0.error}'.format(self)
        super().__init__(self.fmt)


class NotResponding(RequestError):
    """Raised if the API request timed out"""

    def __init__(self):
        self.code = 504
        self.error = 'API request timed out, please be patient.'
        super().__init__(self.error)


class NetworkError(RequestError):
    """Raised if there is an issue with the network
    (i.e. requests.ConnectionError)
    """

    def __init__(self):
        self.code = 503
        self.error = 'Network down.'
        super().__init__(self.error)


class BadRequest(StatusError):
    """Raised when status code 400 is returned.
    Typically when at least one search parameter
    was not provided
    """


class NotFoundError(StatusError):
    """Raised if no result is found"""


class ServerError(StatusError):
    """Raised if the api service is having issues"""


class Unauthorized(StatusError):
    """Raised if you passed invalid credentials."""


class RatelimitError(StatusError):
    """Raised if ratelimit is hit"""


class UnexpectedError(StatusError):
    """Raised if the error was not caught"""


class AuthFlowError(Exception):
    """Raised if no auth method available"""


class NoRefreshToken(Exception):
    """Raised if refresh token is needed but not provided"""


class FileEncryptionError(Exception):
    """Raised if something is wrong with file encryption"""



import pathlib
import re
import sys
from os import system
from setuptools import setup, find_packages


# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    system('python setup.py sdist bdist_wheel')
    system('twine upload dist/*')
    sys.exit()

if sys.version_info < (3, 6, 0):
    raise RuntimeError("audible requires Python 3.6.0+")

here = pathlib.Path(__file__).parent

long_description = (here / 'README.md').read_text('utf-8')

about = (here / 'src' / 'audible' / '_version.py').read_text('utf-8')


def read_from_file(key):
    return re.search(f"{key} = ['\"]([^'\"]+)['\"]", about).group(1)


setup(
    name=read_from_file('__title__'),
    version=read_from_file('__version__'),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    description=read_from_file('__description__'),
    url=read_from_file('__url__'),
    license=read_from_file('__license__'),
    author=read_from_file('__author__'),
    author_email=read_from_file('__author_email__'),
    classifiers=[
         'Development Status :: 4 - Beta',
         'Intended Audience :: Developers',
         'License :: OSI Approved :: GNU Affero General Public License v3',
         'Programming Language :: Python :: 3.6',
         'Programming Language :: Python :: 3.7',
         'Programming Language :: Python :: 3.8'
    ],
    install_requires=[
        'beautifulsoup4',
        'httpx==0.16.*',
        'pbkdf2',
        'Pillow',
        'pyaes',
        'rsa'
    ],
    extras_require={
        'docs': [
            'sphinx',
            'sphinx_rtd_theme',
            'sphinxcontrib-httpdomain',
            'sphinx-autodoc-typehints'
        ]
    },
    python_requires='>=3.6',
    keywords='Audible, API, async',
    long_description=long_description,
    long_description_content_type='text/markdown',
    project_urls={
        "Documentation": "https://audible.readthedocs.io/",
        "Source": "https://github.com/mkb79/Audible",
    }
)



# -*- coding: utf-8 -*-

from ._logging import log_helper
from ._version import __version__
from .auth import Authenticator, LoginAuthenticator, FileAuthenticator
from .client import Client, AsyncClient

__all__ = [
    "__version__", "LoginAuthenticator", "FileAuthenticator", "Authenticator",
    "log_helper", "Client", "AsyncClient"
]



import logging
import pathlib
from typing import Optional, Union
from warnings import warn

logger = logging.getLogger("audible")
logger.addHandler(logging.NullHandler())

log_formatter = logging.Formatter(
    "%(asctime)s %(levelname)s [%(name)s] %(filename)s:%(lineno)d: %(message)s"
)


class AudibleLogHelper:
    def set_level(self, level: Union[str, int]) -> None:
        """Set logging level for the audible package."""
        self._set_level(logger, level)

    @staticmethod
    def _set_level(obj, level: Optional[Union[str, int]]) -> None:
        if level:
            level = level.upper() if isinstance(level, str) else level
            obj.setLevel(level)

        level_name = logging.getLevelName(obj.level)
        logger.info(f"set log level for {obj.name} to: {level_name}")

        if 0 < obj.level < logger.level:
            warn(f"{obj.name} level is lower than {logger.name} logger level")

    def _set_handler(self, handler, name, level):
        handler.setFormatter(log_formatter)
        handler.set_name(name)
        logger.addHandler(handler)
        self._set_level(handler, level)

    def set_console_logger(self,
                           level: Optional[Union[str, int]] = None) -> None:
        """Set up a console logger to the audible package."""
        handler = logging.StreamHandler()
        # noinspection PyTypeChecker
        self._set_handler(handler, "ConsoleLogger", level)

    def set_file_logger(
            self, filename: str, level: Optional[Union[str, int]] = None
    ) -> None:
        """Set up a file logger to the audible package."""
        filename = pathlib.Path(filename)
        handler = logging.FileHandler(filename)
        # noinspection PyTypeChecker
        self._set_handler(handler, "FileLogger", level)

    @staticmethod
    def capture_warnings(status: bool = True) -> None:
        """Lets the logger capture warnings."""
        logging.captureWarnings(status)
        logger.info(
            f"Capture warnings {'activated' if status else 'deactivated'}"
        )


log_helper = AudibleLogHelper()



import base64
import json
import logging
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Generator, Optional, Union
from typing import TYPE_CHECKING

import httpx
import rsa
from httpx import Cookies

from .activation_bytes import get_activation_bytes as get_ab
from .aescipher import AESCipher, detect_file_encryption
from .exceptions import FileEncryptionError, AuthFlowError, NoRefreshToken
from .login import login
from .register import deregister as deregister_
from .register import register as register_
from .utils import test_convert

if TYPE_CHECKING:
    import pathlib
    from .localization import Locale

logger = logging.getLogger("audible.auth")


def refresh_access_token(refresh_token: str, domain: str) -> Dict[str, Any]:
    """Refreshes an access token.

    Args:    
        refresh_token: The refresh token obtained after a device
            registration.
        domain: The top level domain of the requested Amazon server
            (e.g. com).

    Returns:
        A dict with the new access token and expiration timestamp.

    Note:    
        The new access token is valid for 60 minutes.
    """

    body = {
        "app_name": "Audible",
        "app_version": "3.26.1",
        "source_token": refresh_token,
        "requested_token_type": "access_token",
        "source_token_type": "refresh_token"
    }

    resp = httpx.post(f"https://api.amazon.{domain}/auth/token", data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    expires_in_sec = int(resp_dict["expires_in"])
    expires = (
            datetime.utcnow() + timedelta(seconds=expires_in_sec)).timestamp()

    return {"access_token": resp_dict["access_token"], "expires": expires}


def refresh_website_cookies(refresh_token: str,
                            domain: str,
                            cookies_domain: str) -> Dict[str, str]:
    """Fetches website cookies for a specific domain.

    Args:
        refresh_token: The refresh token obtained after a device
            registration.
        domain: The top level domain of the requested Amazon server
            (e.g. com, de, fr).
        cookies_domain: The top level domain scope for the cookies
            (e.g. com, de, fr).
    
    Returns:
        The requested cookies for the Amazon and Audible website for the given
        `cookies_domain` scope.
    """
    url = f"https://www.amazon.{domain}/ap/exchangetoken"

    body = {
        "app_name": "Audible",
        "app_version": "3.26.1",
        "source_token": refresh_token,
        "requested_token_type": "auth_cookies",
        "source_token_type": "refresh_token",
        "domain": f".amazon.{cookies_domain}"
    }

    resp = httpx.post(url, data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    raw_cookies = resp_dict["response"]["tokens"]["cookies"]
    website_cookies = dict()
    for cookies_domain in raw_cookies:
        for cookie in raw_cookies[cookies_domain]:
            website_cookies[cookie["Name"]] = cookie["Value"].replace(
                r'"', r'')

    return website_cookies


def user_profile(access_token: str, domain: str) -> Dict[str, Any]:
    """Returns user profile from Amazon.

    Args:
        access_token: The valid access token for authentication.
        domain: The top level domain of the requested Amazon server
            (e.g. com, de, fr).
    
    Returns:
        The Amazon user profile for the authenticated user.
    """

    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(
        f"https://api.amazon.{domain}/user/profile", headers=headers)
    resp.raise_for_status()

    return resp.json()


def user_profile_audible(access_token: str, domain: str) -> Dict[str, Any]:
    """Returns user profile from Audible.

    Args:
        access_token: The valid access token for authentication.
        domain: The top level domain of the requested Audible server
            (e.g. com, de, fr).
    
    Returns:
        The Audible user profile for the authenticated user.
    """

    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(
        f"https://api.audible.{domain}/user/profile", headers=headers)
    resp.raise_for_status()

    return resp.json()


def sign_request(method: str,
                 path: str,
                 body: bytes,
                 adp_token: str,
                 private_key: str) -> Dict[str, str]:
    """Helper function who creates signed headers for http requests.

    Args:
        path: The requested http url path and query.
        method: The http request method (GET, POST, DELETE, ...).
        body: The http message body.
        adp_token: The adp token obtained after a device registration.
        private_key: The rsa key obtained after device registration.
    
    Returns:
        A dict with the signed headers.
    """
    date = datetime.utcnow().isoformat("T") + "Z"
    body = body.decode("utf-8")

    data = f"{method}\n{path}\n{date}\n{body}\n{adp_token}"

    key = rsa.PrivateKey.load_pkcs1(private_key.encode("utf-8"))
    cipher = rsa.pkcs1.sign(data.encode(), key, "SHA-256")
    signed_encoded = base64.b64encode(cipher)

    signature = f"{signed_encoded.decode()}:{date}"

    return {
        "x-adp-token": adp_token,
        "x-adp-alg": "SHA256withRSA:1.0",
        "x-adp-signature": signature
    }


class Authenticator(httpx.Auth):
    """Audible Authenticator class

    Note:
        A new class instance have to be instantiate with
        :meth:`Authenticator.from_login` or :meth:`Authenticator.from_file`.

    Attributes:
        access_token (:obj:`str`, :obj:`None`):
        adp_token (:obj:`str`, :obj:`None`):
        crypter (:class:`~audible.aescipher.AESCipher`, :obj:`None`):
        customer_info (:obj:`dict`, :obj:`None`):
        device_info (:obj:`dict`, :obj:`None`):
        device_private_key (:obj:`str`, :obj:`None`):
        encryption (:obj:`str`, :obj:`bool`, :obj:`None`):
        expires (:obj:`float`, :obj:`None`):
        filename (:class:`pathlib.Path`, :obj:`None`):
        locale (:class:`~audible.localization.Locale`, :obj:`None`):
        refresh_token (:obj:`str`, :obj:`None`):
        store_authentication_cookie (:obj:`dict`, :obj:`None`):
        website_cookies (:obj:`dict`, :obj:`None`):
    """

    access_token: Optional[str] = None
    adp_token: Optional[str] = None
    crypter: Optional[AESCipher] = None
    customer_info: Optional[Dict] = None
    device_info: Optional[Dict] = None
    device_private_key: Optional[str] = None
    encryption: Optional[Union[str, bool]] = None
    expires: Optional[float] = None
    filename: Optional["pathlib.Path"] = None
    locale: Optional["Locale"] = None
    refresh_token: Optional[str] = None
    store_authentication_cookie: Optional[Dict] = None
    website_cookies: Optional[Dict] = None
    requires_request_body: bool = True
    _forbid_new_attrs: bool = True
    _apply_test_convert: bool = True

    def __setattr__(self, attr, value):
        if self._forbid_new_attrs and not hasattr(self, attr):
            msg = (f"{self.__class__.__name__} is frozen, can't "
                   f"add attribute: {attr}.")
            logger.error(msg)
            raise AttributeError(msg)
                                 
        if self._apply_test_convert:
            value = test_convert(attr, value)
        object.__setattr__(self, attr, value)

    def __iter__(self):
        for i in self.__dict__:
            if self.__dict__[i] is not None and not i.startswith("_"):
                yield i

    def __len__(self):
        return len([i for i in self])

    def __repr__(self):
        return f"{type(self).__name__}({self.__dict__})"

    def _update_attrs(self, **kwargs) -> None:
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @classmethod
    def from_file(cls,
                  filename: Union[str, "pathlib.Path"],
                  password: Optional[str] = None,
                  locale: Optional[Union[str, "Locale"]] = None,
                  encryption: Optional[Union[bool, str]] = None,
                  **kwargs) -> "Authenticator":
        """Instantiate an Authenticator from authentication file.

        .. versionadded:: v0.5.0

        Args:
            filename:
                The name of the file with the authentication data.
            password: The password of the authentication file.
            locale: The country code of the Audible marketplace to interact
                with. If ``None`` the country code from file is used.
            encryption: The encryption style to use. Can be ``json`` or 
                ``bytes``. If ``None``, encryption will be auto detected.
            **kwargs: Keyword arguments are passed to the
                :class:`~audible.aescipher.AESCipher` class. See below.

        Keyword Arguments:
            key_size (int, Optional):
            salt_marker (Optional[bytes]):
            kdf_iterations (:obj:`int`, optional):
            hashmod:
            mac:

        Returns:
            A new Authenticator instance.
        """
        auth = cls()
        auth.filename = filename
        auth.encryption = encryption or detect_file_encryption(auth.filename)

        if auth.encryption:
            if password is None:
                message = "File is encrypted but no password provided."
                logger.critical(message)
                raise FileEncryptionError(message)
            auth.crypter = AESCipher(password, **kwargs)
            file_data = auth.crypter.from_file(auth.filename, auth.encryption)
        else:
            file_data = auth.filename.read_text()

        json_data = json.loads(file_data)

        locale_code = json_data.pop("locale_code", None)
        locale = locale or locale_code
        auth.locale = locale

        # login cookies where renamed to website cookies
        # old names must be adjusted
        login_cookies = json_data.pop("login_cookies", None)
        if login_cookies:
            json_data["website_cookies"] = login_cookies

        auth._update_attrs(**json_data)

        logger.info((f"load data from file {auth.filename} for "
                     f"locale {auth.locale.country_code}"))
        return auth

    @classmethod
    def from_login(cls,
                   username: str,
                   password: str,
                   locale: Union[str, "Locale"],
                   register: bool = False,
                   captcha_callback: Optional[Callable[[str], str]] = None,
                   otp_callback: Optional[Callable[[], str]] = None,
                   cvf_callback: Optional[Callable[[], str]] = None
                   ) -> "Authenticator":
        """Instantiate a new Authenticator with authentication data from login.

        .. versionadded:: v0.5.0

        Args:
            username: The Amazon email address.
            password: The Amazon password.
            locale: The ``country_code`` or :class:`audible.localization.Locale`
                instance for the marketplace to login.
            register: If ``True``, register a new device after login.
            captcha_callback: A custom callback to handle captcha requests
                during login
            otp_callback: A custom callback to handle one-time password
                requests during login.
            cvf_callback: A custom callback to handle verify code requests
                during login.

        Returns:
            An :class:`~audible.auth.Authenticator` instance.
        """
        auth = cls()
        auth.locale = locale

        auth.re_login(
            username=username,
            password=password,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback,
            cvf_callback=cvf_callback)

        logger.info(f"logged in to Audible as {username}")

        if register:
            auth.register_device()
            logger.info("registered Audible device")

        return auth

    def auth_flow(self,
                  request: httpx.Request
                  ) -> Generator[httpx.Request, httpx.Response, None]:
        """Auth flow to be executed on every request by :mod:`httpx`.

        Args:
            request: The request made by ``httpx``.

        Yields:
            The next request
        """
        available_modes = self.available_auth_modes

        if "signing" in available_modes:
            self._apply_signing_auth_flow(request)
        elif "bearer" in available_modes:
            self._apply_bearer_auth_flow(request)
        else:
            message = "signing or bearer auth flow are not available."
            logger.critical(message)
            raise AuthFlowError(message)

        yield request

    def _apply_signing_auth_flow(self, request: httpx.Request) -> None:
        headers = sign_request(method=request.method,
                               path=request.url.raw_path.decode(),
                               body=request.content,
                               adp_token=self.adp_token,
                               private_key=self.device_private_key)

        request.headers.update(headers)
        logger.info("signing auth flow applied to request")

    def _apply_bearer_auth_flow(self, request: httpx.Request) -> None:
        if self.access_token_expired:
            self.refresh_access_token()

        headers = {
            "Authorization": "Bearer " + self.access_token,
            "client-id": "0"
        }
        request.headers.update(headers)
        logger.info("bearer auth flow applied to request")

    def _apply_cookies_auth_flow(self, request: httpx.Request) -> None:
        cookies = {
            name: value
            for (name, value) in self.website_cookies.items()
        }

        Cookies(cookies).set_cookie_header(request)
        logger.info("cookies auth flow applied to request")

    def sign_request(self, request: httpx.Request) -> None:
        """
        .. deprecated:: 0.5.0
           Use :meth:`self._apply_signing_auth_flow` instead.
        """
        self._apply_signing_auth_flow(request)

    @property
    def available_auth_modes(self) -> List:
        available_modes = []

        if self.adp_token and self.device_private_key:
            available_modes.append("signing")

        if self.access_token:
            if self.access_token_expired and not self.refresh_token:
                pass
            else:
                available_modes.append("bearer")

        if self.website_cookies:
            available_modes.append("cookies")

        return available_modes

    def to_file(self,
                filename: Union["pathlib.Path", str] = None,
                password: str = None,
                encryption: Union[bool, str] = "default",
                indent: int = 4,
                set_default: bool = True,
                **kwargs) -> None:

        if not (filename or self.filename):
            raise ValueError("No filename provided")

        if filename:
            filename = test_convert("filename", filename)

        target_file = filename or self.filename

        if encryption != "default":
            encryption = test_convert("encryption", encryption)
        else:
            encryption = self.encryption or False

        body = {
            "website_cookies": self.website_cookies,
            "adp_token": self.adp_token,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "device_private_key": self.device_private_key,
            "store_authentication_cookie": self.store_authentication_cookie,
            "device_info": self.device_info,
            "customer_info": self.customer_info,
            "expires": self.expires,
            "locale_code": self.locale.country_code
        }
        json_body = json.dumps(body, indent=indent)

        if encryption is False:
            target_file.write_text(json_body)
            crypter = None
        else:
            if password:
                crypter = test_convert("crypter", AESCipher(
                    password, **kwargs))
            elif self.crypter:
                crypter = self.crypter
            else:
                raise ValueError("No password provided")

            crypter.to_file(
                json_body,
                filename=target_file,
                encryption=encryption,
                indent=indent)

        logger.info(f"saved data to file {filename}")

        if set_default:
            self.filename = target_file
            self.encryption = encryption
            self.crypter = crypter

        logger.info(f"set filename {target_file} as default")

    def re_login(self,
                 username: str,
                 password: str,
                 captcha_callback: Optional[Callable[[str], str]] = None,
                 otp_callback: Optional[Callable[[], str]] = None,
                 cvf_callback: Optional[Callable[[], str]] = None) -> None:

        login_device = login(
            username=username,
            password=password,
            country_code=self.locale.country_code,
            domain=self.locale.domain,
            market_place_id=self.locale.market_place_id,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback,
            cvf_callback=cvf_callback)

        self._update_attrs(**login_device)

    def register_device(self) -> None:
        register_device = register_(
            access_token=self.access_token, domain=self.locale.domain)

        self._update_attrs(**register_device)

    def deregister_device(self, deregister_all: bool = False) -> Dict[
            str, Any]:
        return deregister_(
            access_token=self.access_token,
            deregister_all=deregister_all,
            domain=self.locale.domain)

    def refresh_access_token(self, force: bool = False) -> None:
        if force or self.access_token_expired:
            if self.refresh_token is None:
                message = "No refresh token found. Can't refresh access token."
                logger.critical(message)
                raise NoRefreshToken(message)
            refresh_data = refresh_access_token(
                refresh_token=self.refresh_token, domain=self.locale.domain)

            self._update_attrs(**refresh_data)
        else:
            logger.info("Access Token not expired. No refresh necessary. "
                        "To force refresh please use force=True")

    def set_website_cookies_for_country(self, country_code: str) -> None:
        cookies_domain = test_convert("locale", country_code).domain

        self.website_cookies = refresh_website_cookies(
            self.refresh_token, self.locale.domain, cookies_domain)

    def get_activation_bytes(self,
                             filename: Optional[
                                 Union["pathlib.Path", str]] = None,
                             extract: bool = True) -> Union[str, bytes]:
        """Get Activation bytes from Audible

        :param filename: [Optional] filename to save the activation blob
        :type filename: string, pathlib.Path
        :param extract: [Optional] if True, returns the extracted activation
                        bytes otherwise the whole activation blob
        :type extract: bool
        """
        return get_ab(self, filename, extract)

    def user_profile(self) -> Dict:
        return user_profile(
            access_token=self.access_token, domain=self.locale.domain)

    @property
    def access_token_expires(self) -> timedelta:
        return datetime.fromtimestamp(self.expires) - datetime.utcnow()

    @property
    def access_token_expired(self) -> bool:
        return datetime.fromtimestamp(self.expires) <= datetime.utcnow()


class LoginAuthenticator:
    """Authenticator class to retrieve credentials from login.
    
    .. deprecated:: v0.5.0
    
       Use :meth:`audible.Authenticator.from_login` instead.
    """

    def __new__(cls,
                username: str,
                password: str,
                locale: Union[str, "Locale"],
                register: bool = False,
                captcha_callback: Optional[Callable[[str], str]] = None,
                otp_callback: Optional[Callable[[], str]] = None,
                cvf_callback: Optional[Callable[[], str]] = None
                ) -> "Authenticator":
        return Authenticator.from_login(username, password, locale, register,
                                        captcha_callback, otp_callback,
                                        cvf_callback)


class FileAuthenticator:
    """Authenticator class to retrieve credentials from stored file.
    
    .. deprecated:: v0.5.0
    
       Use :meth:`audible.Authenticator.from_file` instead.
    """

    def __new__(cls,
                filename: Union[str, "pathlib.Path"],
                password: Optional[str] = None,
                locale: Optional[Union[str, "Locale"]] = None,
                encryption: Optional[Union[bool, str]] = None,
                **kwargs) -> "Authenticator":
        return Authenticator.from_file(filename, password, locale, encryption,
                                       **kwargs)
